﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="Transform3D.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Collision.
//
//   starLiGHT.Collision is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Collision is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Collision. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Collision are available on request.
// </license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 3062                  $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2011-01-06 10:38:34 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.Collision
{
    #region Using Statement
    using System;
    using System.Collections.Generic;
    using System.Text;
    using Microsoft.Xna.Framework;
    #endregion

    public class Transform3D : IEquatable<Transform3D>
    {
        private Matrix transformationMatrix;
        private bool dirty;
        private Matrix originMatrix;
        private bool originMatrixDirty;
        private Matrix scaleMatrix;
        private bool scaleMatrixDirty;
        private Matrix translationMatrix;
        private bool translationMatrixDirty;
        private Matrix rotationMatrix;
        private bool rotationMatrixDirty;

        private Vector3? origin;
        private Vector3? position;
        private Vector3? scale;
        private Vector3? orientation;

        public Transform3D(Vector3 position)
        {
            this.dirty = true;
            this.position = position; 
            this.translationMatrixDirty = true;
            this.origin = null; 
            this.originMatrixDirty = true;
            this.scale = null; 
            this.scaleMatrixDirty = true;
            this.orientation = null; 
            this.rotationMatrixDirty = true;
            
            this.transformationMatrix = Matrix.Identity;
            this.originMatrix = Matrix.Identity;
            this.scaleMatrix = Matrix.Identity;
            this.translationMatrix = Matrix.Identity;
            this.rotationMatrix = Matrix.Identity;
        }

        public Transform3D(Vector3 position, Vector3 orientation)
            : this(position)
        {
            this.orientation = orientation;
        }

        public Transform3D(Vector3 position, Vector3 orientation, float scale)
            : this(position, orientation)
        {
            this.scale = new Vector3(scale);
        }

        public Transform3D(Vector3 position, Vector3 orientation, Vector3 scale)
            : this(position, orientation)
        {
            this.scale = scale;
        }

        public Transform3D(Vector3 position, Vector3 origin, Vector3 orientation, Vector3 scale)
            : this(position, orientation, scale)
        {
            this.origin = origin;
        }

        public Vector3? Origin
        {
            get 
            { 
                return this.origin; 
            }

            set 
            {
                if (this.origin != value)
                {
                    this.originMatrixDirty |= true;
                    this.origin = value;
                }
            }
        }

        public Vector3? Position
        {
            get 
            { 
                return this.position; 
            }

            set 
            {
                if (this.position != value)
                {
                    this.translationMatrixDirty |= true;
                    this.position = value;
                }
            }
        }

        public Vector3? Scale
        {
            get 
            { 
                return this.scale; 
            }

            set
            {
                if (this.scale != value)
                {
                    this.scaleMatrixDirty |= true;
                    this.scale = value;
                }
            }
        }

        public Vector3? Orientation
        {
            get 
            { 
                return this.orientation; 
            }

            set
            {
                if (this.orientation != value)
                {
                    this.rotationMatrixDirty |= true;
                    this.orientation = value;
                }
            }
        }

        public Matrix TransformationMatrix
        {
            get 
            {
                this.UpdateMatrix();
                return this.transformationMatrix; 
            }
        }

        public void Clear()
        {
            this.dirty = true;
            this.origin = null;
            this.originMatrixDirty = true;
            this.position = null;
            this.translationMatrixDirty = true;
            this.scale = null;
            this.scaleMatrixDirty = true;
            this.orientation = null;
            this.rotationMatrixDirty = true;
        }

        public bool Equals(Transform3D other)
        {
            return this.origin.Equals(other.origin) &&
                   this.orientation.Equals(other.orientation) &&
                   this.scale.Equals(other.scale) &&
                   this.position.Equals(other.position);
        }

        private void UpdateMatrix()
        {
            if (this.originMatrixDirty)
            {
                this.originMatrix = this.origin.HasValue ? Matrix.CreateTranslation(-this.origin.Value) : Matrix.Identity;
                this.originMatrixDirty = false;
                this.dirty = true;
            }

            if (this.scaleMatrixDirty)
            {
                this.scaleMatrix = this.scale.HasValue ? Matrix.CreateScale(this.scale.Value) : Matrix.Identity;
                this.scaleMatrixDirty = false;
                this.dirty = true;
            }

            if (this.translationMatrixDirty)
            {
                this.translationMatrix = this.position.HasValue ? Matrix.CreateTranslation(this.position.Value) : Matrix.Identity;
                this.translationMatrixDirty = false;
                this.dirty = true;
            }

            if (this.rotationMatrixDirty)
            {
                this.rotationMatrix = this.orientation.HasValue ? Matrix.CreateFromYawPitchRoll(this.orientation.Value.X, this.orientation.Value.Y, this.orientation.Value.Z) : Matrix.Identity;
                this.rotationMatrixDirty = false;
                this.dirty = true;
            }

            if (this.dirty)
            {
                this.transformationMatrix = this.originMatrix * 
                                            this.scaleMatrix * 
                                            this.rotationMatrix * 
                                            this.translationMatrix;
                this.dirty = false;
            }
        }
    }
}
