#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="Perlin.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2007, 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
//
//    Based on libnoise by Jason Bevins
//      Copyright (C) 2003, 2004 Jason Bevins (licensed under LGPL)
// </copyright>
// <license>
//   This file is part of starLiGHT.Noise.
//
//   starLiGHT.Noise is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Noise is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Noise. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Noise are available on request.
// </license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 10385                 $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2011-06-14 07:59:17 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.Noise.Noise
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;

    #endregion

    /// Noise module that outputs 3-dimensional Perlin noise.
    ///
    /// @image html moduleperlin.png
    ///
    /// Perlin noise is the sum of several coherent-noise functions of
    /// ever-increasing frequencies and ever-decreasing amplitudes.
    ///
    /// An important property of Perlin noise is that a small change in the
    /// input value will produce a small change in the output value, while a
    /// large change in the input value will produce a random change in the
    /// output value.
    ///
    /// This noise module outputs Perlin-noise values that usually range from
    /// -1.0 to +1.0, but there are no guarantees that all output values will
    /// exist within that range.
    ///
    /// For a better description of Perlin noise, see the links in the
    /// <i>References and Acknowledgments</i> section.
    ///
    /// This noise module does not require any source modules.
    ///
    /// <b>Octaves</b>
    ///
    /// The number of octaves control the <i>amount of detail</i> of the
    /// Perlin noise.  Adding more octaves increases the detail of the Perlin
    /// noise, but with the drawback of increasing the calculation time.
    ///
    /// An octave is one of the coherent-noise functions in a series of
    /// coherent-noise functions that are added together to form Perlin
    /// noise.
    ///
    /// An application may specify the frequency of the first octave by
    /// calling the SetFrequency() method.
    ///
    /// An application may specify the number of octaves that generate Perlin
    /// noise by calling the SetOctaveCount() method.
    ///
    /// These coherent-noise functions are called octaves because each octave
    /// has, by default, double the frequency of the previous octave.  Musical
    /// tones have this property as well; a musical C tone that is one octave
    /// higher than the previous C tone has double its frequency.
    ///
    /// <b>Frequency</b>
    ///
    /// An application may specify the frequency of the first octave by
    /// calling the SetFrequency() method.
    ///
    /// <b>Persistence</b>
    ///
    /// The persistence value controls the <i>roughness</i> of the Perlin
    /// noise.  Larger values produce rougher noise.
    ///
    /// The persistence value determines how quickly the amplitudes diminish
    /// for successive octaves.  The amplitude of the first octave is 1.0.
    /// The amplitude of each subsequent octave is equal to the product of the
    /// previous octave's amplitude and the persistence value.  So a
    /// persistence value of 0.5 sets the amplitude of the first octave to
    /// 1.0; the second, 0.5; the third, 0.25; etc.
    ///
    /// An application may specify the persistence value by calling the
    /// SetPersistence() method.
    ///
    /// <b>Lacunarity</b>
    ///
    /// The lacunarity specifies the frequency multipler between successive
    /// octaves.
    ///
    /// The effect of modifying the lacunarity is subtle; you may need to play
    /// with the lacunarity value to determine the effects.  For best results,
    /// set the lacunarity to a number between 1.5 and 3.5.
    ///
    /// <b>References &amp; acknowledgments</b>
    ///
    /// <a href=http://www.noisemachine.com/talk1/>The Noise Machine</a> -
    /// From the master, Ken Perlin himself.  This page contains a
    /// presentation that describes Perlin noise and some of its variants.
    /// He won an Oscar for creating the Perlin noise algorithm!
    ///
    /// <a
    /// href=http://freespace.virgin.net/hugo.elias/models/this.perlin.htm>
    /// Perlin Noise</a> - Hugo Elias's webpage contains a very good
    /// description of Perlin noise and describes its many applications.  This
    /// page gave me the inspiration to create libnoise in the first place.
    /// Now that I know how to generate Perlin noise, I will never again use
    /// cheesy subdivision algorithms to create terrain (unless I absolutely
    /// need the speed.)
    ///
    /// <a
    /// href=http://www.robo-murito.net/code/perlin-noise-math-faq.html>The
    /// Perlin noise math FAQ</a> - A good page that describes Perlin noise in
    /// plain English with only a minor amount of math.  During development of
    /// libnoise, I noticed that my coherent-noise function generated terrain
    /// with some "regularity" to the terrain features.  This page describes a
    /// better coherent-noise function called <i>gradient noise</i>.  This
    /// version of noise::module::Perlin uses gradient coherent noise to
    /// generate Perlin noise.
    public class Perlin : Module
    {
        /// Default frequency for the noise::module::Perlin noise module.
        public const double DEFAULTPERLINFREQUENCY = 1.0;

        /// Default lacunarity for the noise::module::Perlin noise module.
        public const double DEFAULTPERLINLACUNARITY = 2.0;

        /// Default number of octaves for the noise::module::Perlin noise module.
        public const int DEFAULTPERLINOCTAVECOUNT = 6;

        /// Default persistence value for the noise::module::Perlin noise module.
        public const double DEFAULTPERLINPERSISTENCE = 0.5;

        /// Maximum number of octaves for the noise::module::Perlin noise module.
        public const int PERLINMAXOCTAVE = 30;

        /// Frequency of the first octave.
        protected double frequency;

        /// Frequency multiplier between successive octaves.
        protected double lacunarity;

        /// Quality of the Perlin noise.
        protected NoiseQuality noiseQuality;

        /// Total number of octaves that generate the Perlin noise.
        protected int octaveCount;

        /// Persistence of the Perlin noise.
        protected double persistence;

        /// Seed value used by the Perlin-noise function.
        protected int seed;

        /// Default noise quality for the noise::module::Perlin noise module.
        private const NoiseQuality DEFAULT_PERLIN_QUALITY = NoiseQuality.QUALITY_STD;

        /// Default noise seed for the noise::module::Perlin noise module.
        private const int DEFAULT_PERLIN_SEED = 0;

        private List<Parameter> parameterList;

        /// Constructor.
        ///
        /// The default frequency is set to
        /// noise::module::DEFAULT_PERLIN_FREQUENCY.
        ///
        /// The default lacunarity is set to
        /// noise::module::DEFAULT_PERLIN_LACUNARITY.
        ///
        /// The default number of octaves is set to
        /// noise::module::DEFAULT_PERLIN_OCTAVE_COUNT.
        ///
        /// The default persistence value is set to
        /// noise::module::DEFAULT_PERLIN_PERSISTENCE.
        ///
        /// The default seed value is set to
        /// noise::module::DEFAULT_PERLIN_SEED.
        public Perlin() : base(0)
        {
            this.frequency = DEFAULTPERLINFREQUENCY;
            this.lacunarity = DEFAULTPERLINLACUNARITY;
            this.noiseQuality = DEFAULT_PERLIN_QUALITY;
            this.octaveCount = DEFAULTPERLINOCTAVECOUNT;
            this.persistence = DEFAULTPERLINPERSISTENCE;
            this.seed = DEFAULT_PERLIN_SEED;
        }

        public override List<Parameter> ParameterList
        {
            get
            {
                if (this.parameterList == null)
                {
                    this.parameterList = new List<Parameter>();
                    this.parameterList.Add(new Parameter("Seed", "int", DEFAULT_PERLIN_SEED, int.MinValue, int.MaxValue));
                    this.parameterList.Add(new Parameter("Persistence", "double", DEFAULTPERLINPERSISTENCE, 0.0, 1.0));
                    this.parameterList.Add(new Parameter("OctaveCount", "int", DEFAULTPERLINOCTAVECOUNT, 1, PERLINMAXOCTAVE));
                    this.parameterList.Add(new Parameter("NoiseQuality", "enum", NoiseQuality.QUALITY_STD, null, null));
                    this.parameterList.Add(new Parameter("Frequency", "double", DEFAULTPERLINFREQUENCY, 0.0, 16.0));
                    this.parameterList.Add(new Parameter("Lacunarity", "double", DEFAULTPERLINLACUNARITY, 0.0, 5.0));
                }

                return this.parameterList;
            }
        }

        /// Returns the frequency of the first octave.
        ///
        /// @returns The frequency of the first octave.
        public double GetFrequency()
        {
            return this.frequency;
        }

        /// Returns the lacunarity of the Perlin noise.
        ///
        /// @returns The lacunarity of the Perlin noise.
        /// 
        /// The lacunarity is the frequency multiplier between successive
        /// octaves.
        public double GetLacunarity()
        {
            return this.lacunarity;
        }

        /// Returns the quality of the Perlin noise.
        ///
        /// @returns The quality of the Perlin noise.
        ///
        /// See noise::NoiseQuality for definitions of the various
        /// coherent-noise qualities.
        public NoiseQuality GetNoiseQuality()
        {
            return this.noiseQuality;
        }

        /// Returns the number of octaves that generate the Perlin noise.
        ///
        /// @returns The number of octaves that generate the Perlin noise.
        ///
        /// The number of octaves controls the amount of detail in the Perlin
        /// noise.
        public int GetOctaveCount()
        {
            return this.octaveCount;
        }

        /// Returns the persistence value of the Perlin noise.
        ///
        /// @returns The persistence value of the Perlin noise.
        ///
        /// The persistence value controls the roughness of the Perlin noise.
        public double GetPersistence()
        {
            return this.persistence;
        }

        /// Returns the seed value used by the Perlin-noise function.
        ///
        /// @returns The seed value.
        public int GetSeed()
        {
            return this.seed;
        }

        public override double GetValue(double x, double y, double z, double lowerXBound, double upperXBound, double lowerZBound, double upperZBound)
        {
            double value = 0.0;
            double signal = 0.0;
            double curPersistence = 1.0;
            double nx, ny, nz;
            int seed;

            x *= this.frequency;
            y *= this.frequency;
            z *= this.frequency;

            for (int curOctave = 0; curOctave < this.octaveCount; curOctave++)
            {

                // Make sure that these floating-point values have the same range as a 32-
                // bit integer so that we can pass them to the coherent-noise functions.
                nx = noisegen.MakeInt32Range(x);
                ny = noisegen.MakeInt32Range(y);
                nz = noisegen.MakeInt32Range(z);

                // Get the coherent-noise value from the input value and add it to the
                // final result.
                seed = (int)((this.seed + curOctave) & 0xffffffff);
                signal = noisegen.GradientCoherentNoise3D(nx, ny, nz, seed, this.noiseQuality);
                value += signal * curPersistence;

                // Prepare the next octave.
                x *= this.lacunarity;
                y *= this.lacunarity;
                z *= this.lacunarity;
                curPersistence *= this.persistence;
            }

            return value;
        }

        /// Sets the frequency of the first octave.
        ///
        /// @param frequency The frequency of the first octave.
        public void SetFrequency(double frequency)
        {
            this.frequency = frequency;
        }

        /// Sets the lacunarity of the Perlin noise.
        ///
        /// @param lacunarity The lacunarity of the Perlin noise.
        /// 
        /// The lacunarity is the frequency multiplier between successive
        /// octaves.
        ///
        /// For best results, set the lacunarity to a number between 1.5 and
        /// 3.5.
        public void SetLacunarity(double lacunarity)
        {
            this.lacunarity = lacunarity;
        }

        /// Sets the quality of the Perlin noise.
        ///
        /// @param noiseQuality The quality of the Perlin noise.
        ///
        /// See noise::NoiseQuality for definitions of the various
        /// coherent-noise qualities.
        public void SetNoiseQuality(NoiseQuality noiseQuality)
        {
            this.noiseQuality = noiseQuality;
        }

        /// Sets the number of octaves that generate the Perlin noise.
        ///
        /// @param octaveCount The number of octaves that generate the Perlin
        /// noise.
        ///
        /// @pre The number of octaves ranges from 1 to
        /// noise::module::PERLIN_MAX_OCTAVE.
        ///
        /// @throw noise::ExceptionInvalidParam An invalid parameter was
        /// specified; see the preconditions for more information.
        ///
        /// The number of octaves controls the amount of detail in the Perlin
        /// noise.
        ///
        /// The larger the number of octaves, the more time required to
        /// calculate the Perlin-noise value.
        public void SetOctaveCount(int octaveCount)
        {
            if (octaveCount < 1 || octaveCount > PERLINMAXOCTAVE)
                throw new Exception("invalid params");
            this.octaveCount = octaveCount;
        }

        /// Sets the persistence value of the Perlin noise.
        ///
        /// @param persistence The persistence value of the Perlin noise.
        ///
        /// The persistence value controls the roughness of the Perlin noise.
        ///
        /// For best results, set the persistence to a number between 0.0 and
        /// 1.0.
        public void SetPersistence(double persistence)
        {
            this.persistence = persistence;
        }

        /// Sets the seed value used by the Perlin-noise function.
        ///
        /// @param seed The seed value.
        public void SetSeed(int seed)
        {
            this.seed = seed;
        }

        public override void SetParameter(string name, object value)
        {
            switch (name.ToLower())
            {
                case "seed":
                    this.SetSeed(Convert.ToInt32(value));
                    break;
                case "persistence":
                    this.SetPersistence(Convert.ToDouble(value));
                    break;
                case "octavecount":
                    this.SetOctaveCount(Convert.ToInt32(value));
                    break;
                case "noisequality":
                    this.SetNoiseQuality((NoiseQuality)value);
                    break;
                case "frequency":
                    this.SetFrequency(Convert.ToDouble(value));
                    break;
                case "lacunarity":
                    this.SetLacunarity(Convert.ToDouble(value));
                    break;
            }
        }

        public override object GetParameter(string name)
        {
            switch (name.ToLower())
            {
                case "seed": return this.GetSeed();
                case "persistence": return this.GetPersistence();
                case "octavecount": return this.GetOctaveCount();
                case "noisequality": return this.GetNoiseQuality();
                case "frequency": return this.GetFrequency();
                case "lacunarity": return this.GetLacunarity();
            }

            return null;
        }
    }
}

