﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="Age.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Modifier
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.ParticleSystem.Common;
    using starLiGHT.ParticleSystem.Common.EnergyEasing;

    #endregion

    /// <summary>
    /// The Age modifier operates in conjunction with the Lifetime 
    /// initializer. The Lifetime initializer sets the lifetime for 
    /// the particle. The Age modifier then ages the particle over time, 
    /// altering its energy to reflect its age. This energy can then 
    /// be used by modifiers like Fade and ChangeColor to alter the 
    /// appearence of the particle as it ages.
    /// <p>The aging process need not to be linear. This modifier can use 
    /// a function that modifies the aging process, to ease in or out or 
    /// to alter the aging in other ways.</p>
    /// <p>When the particle's lifetime is over, this modifier marks it 
    /// as dead.</p>
    /// <p>When adjusting the energy this modifier can use any of the 
    /// easing functions in the Easing namespace along with any custom 
    /// easing functions that have the same interface.</p>
    /// </summary>
    public class Age : ModifierBase
    {
        #region Member Variables
        private System.Func<float, float, float> easing;

        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the Age class to be used as a modifier by an emitter. 
        /// </summary>
        /// <remarks>
        /// This uses a linear response with no easing.
        /// </remarks>
        public Age()
            : this(null)
        {
        }

        /// <summary>
        /// The constructor creates an Age modifier to be used by an emitter. 
        /// </summary>
        /// <param name="easing">
        /// An easing function to use to modify the  energy curve over the lifetime 
        /// of the particle. A null value produces a linear response with no easing.
        /// </param>
        public Age(System.Func<float, float, float> easing)
        {
            if (easing == null)
            {
                this.easing = starLiGHT.ParticleSystem.Common.EnergyEasing.Linear.easeNone;
            }
            else
            {
                this.easing = easing;
            }
        }

        #endregion

        /// <summary>
        /// An easing function to use to modify the  energy curve over the lifetime 
        /// of the particle. A null value produces a linear response with no easing.
        /// </summary>
        public System.Func<float, float, float> Easing
        {
            get { return this.easing; }
            set { this.easing = value; }
        }

        /// <summary>
        /// Sets the energy of the particle based on its age and the easing function.
        /// <p>This method is called by the emitter and need not be called by the user</p>
        /// </summary>
        /// <param name="emitter">The Emitter that created the particle.</param>
        /// <param name="particle">The particle to be updated.</param>
        /// <param name="time">The duration of the frame - used for time based updates.</param>
        public override void Update(Emitter emitter, Particle particle, float time)
        {
            particle.age += time;

            if (particle.age >= particle.lifetime)
            {
                particle.energy = 0;
                particle.isDead = true;
            }
            else
            {
                particle.energy = this.easing(particle.age, particle.lifetime);
            }
        }
    }
}
