﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="TargetScale.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Modifier
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.ParticleSystem.Common;

    #endregion

    /// <summary>
    /// The TargetScale modifier scales the size of a particle towards a target size. On 
    /// every update the size of the particle moves a little closer to the target size. The 
    /// rate at which particles approach this size is controlled by the rate property.
    /// </summary>
    public class TargetScale : ModifierBase
    {
        #region Member Variables
        private float _scale;
        private float _rate;

        #endregion

        #region Constructors
        /// <summary>
        /// The constructor creates a TargetScale modifier to be used by an emitter.
        /// </summary>
        public TargetScale()
            : this(1.0f, 0.1f)
        {
        }

        /// <summary>
        /// The constructor creates a TargetScale modifier to be used by an emitter.
        /// </summary>
        /// <param name="targetScale">
        /// The target scale that should be reached
        /// </param>
        /// <param name="rate">
        /// Adjusts how quickly the particle approaches the target scale. Larger numbers 
        /// cause a higher speed.
        /// </param>
        public TargetScale(float targetScale, float rate)
        {
            _scale = targetScale;
            _rate = rate;
        }

        #endregion

        /// <summary>
        /// The target scale that should be reached
        /// </summary>
        public float Target
        {
            get { return _scale; }
            set { _scale = value; }
        }

        /// <summary>
        /// Adjusts how quickly the particle approaches the target scale. Larger numbers 
        /// cause a higher speed.
        /// </summary>
        public float Rate
        {
            get { return _rate; }
            set { _rate = value; }
        }

        /// <summary>
        /// Moves the actual scale a little closer to the target scale taking the rate and the 
        /// time since last update into account.
        /// <p>This method is called by the emitter and need not be called by the user</p>
        /// </summary>
        /// <param name="emitter">The Emitter that created the particle.</param>
        /// <param name="particle">The particle to be updated.</param>
        /// <param name="time">The duration of the frame - used for time based updates.</param>
        public override void Update(Emitter emitter, Particle particle, float time)
        {
            particle.scale += (_scale - particle.scale) * _rate * time;
        }
    }
}
