﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="PerformanceAdjusted.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.PulseGenerators
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.Collections;
    using starLiGHT.ParticleSystem.Common;

    #endregion

    /// <summary>
    /// The PerformanceAdjusted PulseGenerator causes the emitter to emit particles 
    /// continuously at a steady rate, but it adjusts this rate dynamically if a given 
    /// frame rate could not be hold. It will adjust the emission rate to a level at which 
    /// the system can maintain the target frame rate.
    /// </summary>
    /// <remarks>
    /// The PerformanceAdjusted PulseGenerator does not make sense to be used on platforms that 
    /// are running at a predictable speed, e.g. the XBox360. The performance measurement is not 
    /// free and on such platforms it is better to set a particle maximum that works in all situations.
    /// </remarks>
    public class PerformanceAdjusted : IPulseGenerator
    {
        #region Member Variables
        private float _timeToNext;
        private uint _rateMin;
        private uint _rateMax;
        private float _target;
        private float _rate;
        private CircularBuffer<float> _times;
        private float _timeToRateCheck;
        private bool _stop;

        #endregion

        #region Constructors
        public PerformanceAdjusted()
            : this(0, 0, 60)
        {
        }

        /// <summary>
        /// The constructor creates a PerformanceAdjusted PulseGenerator to be used by an emitter.
        /// </summary>
        /// <param name="rateMin">
        /// The minimum number of particles to emit per second. The PulseGenerator will never drop 
        /// the pulse below this value, even if the system can not maintain the target frame rate.
        /// </param>
        /// <param name="rateMax">
        /// The maximum number of particles to emit per second. The PulseGenerator will start with 
        /// this pulse rate.
        /// </param>
        /// <param name="targetFrameRate">
        /// The target frame rate, that the PerformanceAdjusted PulseGenerator tries to maintain.
        /// </param>
        public PerformanceAdjusted(uint rateMin, uint rateMax, float targetFrameRate)
        {
            _stop = false;
            _rateMin = rateMin;
            _rate = _rateMax = rateMax;
            _target = targetFrameRate;
            _times = new CircularBuffer<float>(9);
            _timeToRateCheck = 0;
        }

        #endregion

        /// <summary>
        /// The minimum number of particles to emit per second. The PulseGenerator will never drop 
        /// the pulse below this value, even if the system can not maintain the target frame rate.
        /// </summary>
        public uint RateMin
        {
            get 
            { 
                return _rateMin; 
            }

            set
            {
                _rateMin = value;
                _timeToRateCheck = 0;
            }
        }

        /// <summary>
        /// The maximum number of particles to emit per second. The PulseGenerator will start with 
        /// this pulse rate.
        /// </summary>
        public uint RateMax
        {
            get 
            { 
                return _rateMax; 
            }

            set
            {
                _rate = _rateMax = value;
                _timeToRateCheck = 0;
            }
        }

        /// <summary>
        /// The target frame rate, that the PerformanceAdjusted PulseGenerator tries to maintain.
        /// </summary>
        public float TargetFrameRate
        {
            get { return _target; }
            set { _target = value; }
        }

        /// <summary>
        /// The startEmitter method is called when the emitter starts.
        /// </summary>
        /// <remarks>
        /// This method is private and called by the emitter's start method.
        /// </remarks>
        /// <returns>
        /// The number of particles the emitter should emit when it starts.
        /// </returns>
        public uint startEmitter(Emitter emitter)
        {
            newTimeToNext();
            return 0;
        }

       /// <summary>
        /// The updateEmitter method is called every frame after the emitter has been started.
        /// </summary>
        /// <remarks>
        /// This method is private and called by the emitter's update loop.
        /// </remarks>
        /// <param name="time">The time in seconds since the previous call of this method.</param>
        /// <returns>The number of particles the emitter should emit at this time.</returns>
        public uint updateEmitter(Emitter emitter, float time)
        {
            if (_stop)
            {
                return 0;
            }

            if (_rate > _rateMin && (_timeToRateCheck -= time) <= 0)
            {
                // adjust rate
                float t = time;
                _times.Put(t);
                if (_times.Count > 9)
                {
                    float frameRate = (float)Math.Round(10000 / (t - _times[8]));
                    if (frameRate < _target)
                    {
                        _rate = (float)Math.Floor((_rate + _rateMin) * 0.5f);

                        if (!(_timeToRateCheck == emitter.Particles[0].lifetime))
                        {
                            _timeToRateCheck = 2;
                        }
                    }
                }
            }

            float emitTime = time;
            uint count = 0;
            emitTime -= _timeToNext;
            while (emitTime >= 0)
            {
                ++count;
                newTimeToNext();
                emitTime -= _timeToNext;
            }

            _timeToNext = -emitTime;

            return count;
        }

        /// <summary>
        /// Stops the PulseGenerator instructing the emitter to emit particles
        /// </summary>
        public void stop()
        {
            _stop = true;
        }

        /// <summary>
        /// Resumes the PulseGenerator after a stop.
        /// </summary>
        public void resume()
        {
            _stop = false;
        }

        /// <summary>
        /// Indicates, that the PulseGenerator has raised all it's pulses and no more 
        /// particles need to be emitted.
        /// </summary>
        public bool Complete
        {
            get { return false; }
        }

        #region Internal members
        private void newTimeToNext()
        {
            _timeToNext = 1 / _rate;
        }

        #endregion
    }
}
