﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="Sine.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.PulseGenerators
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.ParticleSystem.Common;

    #endregion

    /// <summary>
    /// The Sine PulseGenerator causes the emitter to emit particles continuously 
    /// at a rate that varies according to a sine wave.
    /// </summary>
    public class Sine : IPulseGenerator
    {
        #region Member Variables
        private uint _emitted;
        private float _rateMin;
        private float _rateMax;
        private float _period;
        private bool _stop;
        private float _timePassed;
        private double _factor;
        private float _scale;

        #endregion

        #region Constructors
        
        /// <summary>
        /// The constructor creates a Sine PulseGenerator to be used by an emitter.
        /// </summary>
        public Sine()
            : this(1, 0, 0)
        {
        }

        /// <summary>
        /// The constructor creates a Sine PulseGenerator to be used by an emitter.
        /// </summary>
        /// <param name="period">
        /// The period of the sine wave used, in seconds. This is the time between 
        /// one maximum and one minimum.
        /// </param>
        /// <param name="rateMin">
        /// The number of particles to emit per second at the bottom of the sine wave.
        /// </param>
        /// <param name="rateMax">
        /// The number of particles to emit per second at the peak of the sine wave.
        /// </param>
        public Sine(float period, float rateMin, float rateMax)
        {
            this._stop = false;
            this._period = period;
            this._rateMin = rateMin;
            this._rateMax = rateMax;
            this._factor = 2 * Math.PI / period;
            this._scale = 0.5f * (this._rateMax - this._rateMin);
        }

        #endregion

        /// <summary>
        /// The number of particles to emit per second at the bottom of the sine wave.
        /// </summary>
        public float RateMin
        {
            get 
            {
                return this._rateMin; 
            }

            set
            {
                this._rateMin = value;
                this._scale = 0.5f * (this._rateMax - this._rateMin);
            }
        }

        /// <summary>
        /// The number of particles to emit per second at the peak of the sine wave.
        /// </summary>
        public float RateMax
        {
            get 
            {
                return this._rateMax; 
            }

            set
            {
                this._rateMax = value;
                this._scale = 0.5f * (this._rateMax - this._rateMin);
            }
        }

        /// <summary>
        /// The period of the sine wave used, in seconds. This is the time between 
        /// one maximum and one minimum.
        /// </summary>
        public float Period
        {
            get 
            {
                return this._period; 
            }

            set
            {
                this._period = value;
                this._factor = 2 * Math.PI / this._period;
            }
        }

        /// <summary>
        /// The startEmitter method is called when the emitter starts.
        /// </summary>
        /// <remarks>
        /// This method is private and called by the emitter's start method.
        /// </remarks>
        /// <returns>
        /// The number of particles the emitter should emit when it starts.
        /// </returns>
        public uint startEmitter(Emitter emitter)
        {
            this._timePassed = 0;
            this._emitted = 0;
            return 0;
        }

        /// <summary>
        /// The updateEmitter method is called every frame after the emitter has been started.
        /// </summary>
        /// <remarks>
        /// This method is private and called by the emitter's update loop.
        /// </remarks>
        /// <param name="time">The time in seconds since the previous call of this method.</param>
        /// <returns>The number of particles the emitter should emit at this time.</returns>
        public uint updateEmitter(Emitter emitter, float time)
        {
            if (_stop)
            {
                return 0;
            }

            this._timePassed += time;
            uint count = (uint)Math.Floor(this._rateMax * this._timePassed + this._scale * (1 - Math.Cos(this._timePassed * _factor)) / _factor);
            uint ret = count - this._emitted;
            this._emitted = count;
            return ret;
        }

        /// <summary>
        /// Stops the PulseGenerator instructing the emitter to emit particles
        /// </summary>
        public void stop()
        {
            this._stop = true;
        }

        /// <summary>
        /// Resumes the PulseGenerator after a stop.
        /// </summary>
        public void resume()
        {
            this._stop = false;
            this._emitted = 0;
        }

        /// <summary>
        /// Indicates, that the PulseGenerator has raised all it's pulses and no more 
        /// particles need to be emitted.
        /// </summary>
        public bool Complete
        {
            get { return false; }
        }
    }
}
