﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="TimePeriod.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.PulseGenerators
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.ParticleSystem.Common;

    #endregion

    /// <summary>
    /// The TimePeriod PulseGenerator causes the emitter to emit particles for a period of 
    /// time and then stop. The rate of emission over that period can be modified 
    /// using easing equations that conform to the interface defined in Robert Penner's easing 
    /// equations.
    /// </summary>
    public class TimePeriod : IPulseGenerator
    {
        #region Member Variables
        private uint _particles;
        private float _duration;
        private uint _particlesPassed;
        private float _timePassed;
        private Func<float, float, float, float, float> _easing;
        private bool _stop;

        #endregion

        #region Constructors
        /// <summary>
        /// The constructor creates a TimePeriod PulseGenerator to be used by an emitter.
        /// </summary>
        public TimePeriod()
            : this(0, 0, null)
        {
        }

        /// <summary>
        /// The constructor creates a TimePeriod PulseGenerator to be used by an emitter.
        /// </summary>
        /// <param name="numParticles">
        /// The number of particles to emit over the full duration of the time period.
        /// </param>
        /// <param name="duration">
        /// The duration of the time period. After this time the emitter will stop releasing particles.
        /// </param>
        /// <param name="easing">
        /// An easing function used to distribute the emission of the particles over the time period. If 
        /// no easing function is passed a simple linear distribution is used in which particles are emitted 
        /// at a steady rate over the time period.
        /// </param>
        public TimePeriod(uint numParticles, float duration, Func<float, float, float, float, float> easing)
        {
            this._particles = numParticles;
            this._duration = duration;
            if (easing == null)
            {
                this._easing = Linear.easeNone;
            }
            else
            {
                this._easing = easing;
            }
        }

        #endregion

        /// <summary>
        /// The number of particles to emit over the full duration of the time period.
        /// </summary>
        public uint NumParticles
        {
            get { return this._particles; }
            set { this._particles = value; }
        }

        /// <summary>
        /// The duration of the time period. After this time the emitter will stop releasing particles.
        /// </summary>
        public float Duration
        {
            get { return this._duration; }
            set { this._duration = value; }
        }

        /// <summary>
        /// An easing function used to distribute the emission of the particles over the time period. If 
        /// no easing function is passed a simple linear distribution is used in which particles are emitted 
        /// at a steady rate over the time period.
        /// </summary>
        public Func<float, float, float, float, float> Easing
        {
            get { return this._easing; }
            set { this._easing = value; }
        }

        /// <summary>
        /// The startEmitter method is called when the emitter starts.
        /// </summary>
        /// <remarks>
        /// This method is private and called by the emitter's start method.
        /// </remarks>
        /// <returns>
        /// The number of particles the emitter should emit when it starts.
        /// </returns>
        public uint startEmitter(Emitter emitter)
        {
            this._particlesPassed = 0;
            this._timePassed = 0;
            return 0;
        }

        /// <summary>
        /// The updateEmitter method is called every frame after the emitter has been started.
        /// </summary>
        /// <remarks>
        /// This method is private and called by the emitter's update loop.
        /// </remarks>
        /// <param name="time">The time in seconds since the previous call of this method.</param>
        /// <returns>The number of particles the emitter should emit at this time.</returns>
        public uint updateEmitter(Emitter emitter, float time)
        {
            if (this._stop || this._timePassed >= this._duration)
            {
                return 0;
            }

            this._timePassed += time;

            if (this._timePassed >= this._duration)
            {
                emitter.dispatchPulseGeneratorComplete();
                uint newParticles = this._particles - this._particlesPassed;
                this._particlesPassed = this._particles;
                return newParticles;
            }

            uint oldParticles = this._particlesPassed;
            this._particlesPassed = (uint)Math.Round(this._easing(this._timePassed, 0, this._particles, this._duration));
            return this._particlesPassed - oldParticles;
        }

        /// <summary>
        /// Stops the PulseGenerator instructing the emitter to emit particles
        /// </summary>
        public void stop()
        {
            this._stop = true;
        }

        /// <summary>
        /// Resumes the PulseGenerator after a stop.
        /// </summary>
        public void resume()
        {
            this._stop = false;
        }

        /// <summary>
        /// Indicates, that the PulseGenerator has raised all it's pulses and no more 
        /// particles need to be emitted.
        /// </summary>
        public bool Complete
        {
            get { return this._particlesPassed == this._particles; }
        }
    }
}
