﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="ActivityBase.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Common
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;

    #endregion

    /// <summary>
    /// The ActivityBase class is the abstract base class for all emitter activities
    /// in the Flint library. It implements the Activity interface with a default
    /// priority of zero and empty methods for the rest of the interface.
    /// 
    /// <p>Instances of the ActivityBase class should not be directly created 
    /// because the ActivityBase class itself simply implements the Activity 
    /// interface with default methods that do nothing.
    /// 
    /// <p>Developers creating custom activities may either extend the ActivityBase
    /// class or implement the Activity interface directly. Classes that extend the 
    /// ActivityBase class need only to implement their own functionality for the 
    /// methods they want to use, leaving other methods with their default empty
    /// implementations.
    /// </summary>
    public class ActivityBase : IActivity
    {
        private int _priority;

        /// <summary>
        /// The constructor creates an ActivityBase object. But you shouldn't use it 
        /// directly because the ActivityBase class is abstract.
        /// </summary>
        public ActivityBase()
        {
        }

        /// <summary>
        /// Returns a default priority of 0 for this activity. Derived classes 
        /// override this method if they want a different default priority.
        /// </summary>
        public int Priority
        {
            get { return _priority; }
            set { _priority = value; }
        }

        /// <summary>
        /// This method does nothing. Some derived classes override this method
        /// to perform actions when the activity is added to an emitter.
        /// </summary>
        /// <param name="emitter">
        /// The Emitter that the Activity was added to.
        /// </param>
        public void addedToEmitter(Emitter emitter)
        {
        }

        /// <summary>
        /// This method does nothing. Some derived classes override this method
        /// to perform actions when the activity is removed from the emitter.
        /// </summary>
        /// <param name="emitter">The Emitter that the Activity was removed from.</param>
        public void removedFromEmitter(Emitter emitter)
        {
        }

        /// <summary>
        /// This method does nothing. Derived classes override this method 
        /// to alter the state of the emitter when it starts.
        /// </summary>
        /// <param name="emitter">The Emitter.</param>
        public void initialize(Emitter emitter)
        {
        }

        /// <summary>
        /// This method does nothing. Derived classes override this method 
        /// to alter the state of the emitter every frame.
        /// </summary>
        /// <param name="emitter">The Emitter.</param>
        /// <param name="time">The duration of the frame - used for time based updates.</param>
        public virtual void update(Emitter emitter, float time)
        {
        }
    }
}
