﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="Alpha.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Initializer
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.ParticleSystem.Common;

    #endregion

    /// <summary>
    /// The AlphaInit Initializer sets the alpha transparency of the particle.
    /// </summary>
    public class Alpha : InitializerBase
    {
        private float _min;
        private float _max;

        /// <summary>
        /// The constructor creates an AlphaInit initializer for use by 
        /// an emitter. To add an AlphaInit to all particles created by an emitter, use the
        /// emitter's addInitializer method.
        /// <p>The alpha of particles initialized by this class
        /// will be a random value between the minimum and maximum
        /// values set. If no maximum value is set, the minimum value
        /// is used with no variation.</p>
        /// <p>This initializer has a priority of -10 so that it occurs after 
        /// the color assignment.</p>
        /// </summary>
        public Alpha()
            : this(1, float.NaN)
        {
        }

        /// <summary>
        /// The constructor creates an AlphaInit initializer for use by 
        /// an emitter. To add an AlphaInit to all particles created by an emitter, use the
        /// emitter's addInitializer method.
        /// <para />
        /// <p>The alpha of particles initialized by this class
        /// will be a random value between the minimum and maximum
        /// values set. If no maximum value is set, the minimum value
        /// is used with no variation.</p>
        /// <para />
        /// <p>This initializer has a priority of -10 so that it occurs after 
        /// the color assignment.</p>
        /// </summary>
        public Alpha(float minAlpha)
            : this(minAlpha, float.NaN)
        {
        }

        /// <summary>
        /// The constructor creates an AlphaInit initializer for use by 
        /// an emitter. To add an AlphaInit to all particles created by an emitter, use the
        /// emitter's addInitializer method.
        /// <para />
        /// <p>The alpha of particles initialized by this class
        /// will be a random value between the minimum and maximum
        /// values set. If no maximum value is set, the minimum value
        /// is used with no variation.</p>
        /// <p>This initializer has a priority of -10 so that it occurs after 
        /// the color assignment.</p>
        /// </summary>
        /// <param name="minAlpha">the minimum alpha for particles initialized by the instance. The value should be between 1 and 0.</param>
        /// <param name="maxAlpha">the maximum alpha for particles initialized by the instance. The value should be between 1 and 0.</param>
        public Alpha(float minAlpha, float maxAlpha)
        {
            priority = -10;
            _min = minAlpha;
            if (float.IsNaN(maxAlpha))
            {
                _max = _min;
            }
            else
            {
                _max = maxAlpha;
            }
        }

        /// <summary>
        /// The minimum alpha value for particles initialised by 
        /// this initializer. Should be between 0 and 1.
        /// </summary>
        public float MinAlpha
        {
            get { return _min; }
            set { _min = value; }
        }

        /// <summary>
        /// The maximum alpha value for particles initialised by 
        /// this initializer. Should be between 0 and 1.
        /// </summary>
        public float MaxAlpha
        {
            get { return _max; }
            set { _max = value; }
        }

        /// <summary>
        /// When reading, returns the average of minAlpha and maxAlpha. 
        /// When writing this sets both maxAlpha and minAlpha to the  
        /// same alpha value.
        /// </summary>
        public float AlphaValue
        {
            get { return _min == _max ? _min : (_max + _min) * 0.5f; }
            set { _max = _min = value; }
        }

        /// <summary>
        /// The initialize method is used by the emitter to apply the initialization
        /// to every particle. It is the key feature of the initializers and is
        /// used to initialize the state of every particle. This method 
        /// is called within the emitter's createParticle met6hod for every particle 
        /// and need not be called by the user.
        /// </summary>
        /// <param name="emitter">The Emitter that created the particle.</param>
        /// <param name="particle">The particle to be initialized.</param>
        /// <remarks>This is an internal method that is called by the emitter.</remarks>
        public override void initialize(Emitter emitter, ref Particle particle)
        {
            float alpha;
            if (_max == _min)
            {
                alpha = _min;
            }
            else
            {
                alpha = emitter.Random.NextFloat(_min, _max);
            }

            particle.color.A = (byte)(255 * alpha);
        }
    }
}
