﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="RendererBase.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Common
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;

    #endregion

    /// <summary>
    /// The base class used by all the renderers. This class manages 
    /// various aspects of the rendering process.
    /// <p>The class will add every emitter it should renderer to it's internal 
    /// array of emitters. It will listen for the appropriate events on the 
    /// emitter and will then call the protected methods addParticle, removeParticle 
    /// and renderParticles at the appropriate times. Many derived classes need 
    /// only implement these three methods to manage the rendering of the particles.</p>
    /// </summary>
    public abstract class RendererBase : IRenderer
    {
        private Dictionary<string, Emitter> _emitters;

        /// <summary>
        /// The constructor creates a RendererBase class.
        /// </summary>
        public RendererBase()
        {
            this._emitters = new Dictionary<string, Emitter>();
        }

        public void addEmitter(Emitter emitter)
        {
            this._emitters.Add(emitter.Name, emitter);
            ////////TODO: wire up events
            foreach (Particle p in emitter.Particles)
            {
                addParticle(p);
            }
        }

        public void removeEmitter(Emitter emitter)
        {
            if (this._emitters.ContainsKey(emitter.Name))
            {
                this._emitters.Remove(emitter.Name);
                ////////TODO: unwire events
                foreach (Particle p in emitter.Particles)
                {
                    removeParticle(p);
                }
            }
        }

        public int NumEmitter
        {
            get { return this._emitters.Count; }
        }

        public Dictionary<string, Emitter> Emitters
        {
            get { return this._emitters; }
        }

        /// <summary>
        /// The addParticle method is called when a particle is added to one of 
        /// the emitters that is being rendered by this renderer.
        /// </summary>
        /// <param name="particle">The particle.</param>
        protected abstract void addParticle(Particle particle);

        /// <summary>
        /// The removeParticle method is called when a particle is removed from one 
        /// of the emitters that is being rendered by this renderer.
        /// </summary>
        /// <param name="particle">The particle.</param>
        protected abstract void removeParticle(Particle particle);

        /// <summary>
        /// The renderParticles method is called during the render phase of 
        /// every frame if the state of one of the emitters being rendered 
        /// by this renderer has changed.
        /// </summary>
        /// <param name="particles">
        /// The particles being managed by all the emitters 
        /// being rendered by this renderer. The particles are in no particular 
        /// order.
        /// </param>
        protected abstract void renderParticles(List<Particle> particles);

        #region Renderer Member

        public string Type
        {
            get { throw new NotImplementedException(); }
        }

        #endregion
    }
}
