﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="Position.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 7773                  $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2011-03-28 02:04:04 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Initializer.TwoD
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using Microsoft.Xna.Framework;
    using starLiGHT.ParticleSystem.Common;
    using starLiGHT.ParticleSystem.Initializer;
    using starLiGHT.ParticleSystem.twoD;

    #endregion

    /// <summary>
    /// The Position Initializer sets the initial location of the particle
    /// </summary>
    /// <remarks>
    /// This class is using regions to place the particles. A region defines a 
    /// space in relation to the emitter and the particles are placed at a random 
    /// point within that space. For precise placement, the Point Region defines a 
    /// single point at which all particles will be placed. Various regions (and the
    /// regions interface for use when implementing custom regions) are defined.
    /// </remarks>
    public class Position : InitializerBase
    {
        #region Member Variables
        private IRegion2D _region;
        private float? lastrotation;
        private float sin;
        private float cos;

        #endregion

        #region Constructors
        /// <summary>
        /// The constructor creates a Position Initializer to be used by an emitter.
        /// </summary>
        /// <remarks>
        /// This can be used to add a position offset to all particles created by an emitter.
        /// </remarks>
        public Position()
            : this(null)
        {
        }

        /// <summary>
        /// The constructor creates a Position Initializer to be used by an emitter.
        /// </summary>
        /// <param name="region">
        /// Region to initialize all particles in.
        /// </param>
        /// <remarks>
        /// This can be used to add a position offset to all particles created by an emitter.
        /// </remarks>
        public Position(IRegion2D region)
        {
            _region = region;
        }

        /// <summary>
        /// The constructor creates a Position Initializer to be used by an emitter.
        /// </summary>
        /// <param name="location">
        /// The location is used to create a Point Region for that position.
        /// </param>
        /// <remarks>
        /// This can be used to add a position offset to all particles created by an emitter.
        /// </remarks>
        public Position(Vector2 location)
            : this(new PointRegion(location))
        {
        }

        public Position(float x, float y)
            : this(new PointRegion(x, y))
        {
        }

        #endregion

        /// <summary>
        /// Region to initialize all particles in.
        /// </summary>
        public IRegion2D Region
        {
            get { return _region; }
            set { _region = value; }
        }

        /// <summary>
        /// The initialize method is used by the emitter to apply the initialization
        /// to every particle. It is the key feature of the initializers and is
        /// used to initialize the state of every particle. This method 
        /// is called within the emitter's createParticle met6hod for every particle 
        /// and need not be called by the user.
        /// </summary>
        /// <param name="emitter">The Emitter that created the particle.</param>
        /// <param name="particle">The particle to be initialized.</param>
        /// <remarks>This is an internal method that is called by the emitter.</remarks>
        public override void initialize(Emitter emitter, ref Particle particle)
        {
            Particle2D p = (Particle2D)particle;
            if (_region != null)
            {
                Vector2 loc = _region.getLocation();
                if (p.rotation == 0)
                {
                    p.position.X += loc.X;
                    p.position.Y += loc.Y;
                }
                else if (lastrotation.HasValue && lastrotation.Value == p.rotation)
                {
                    p.position.X += cos * loc.X - sin * loc.Y;
                    p.position.Y += cos * loc.Y + sin * loc.X;
                }
                else
                {
                    sin = (float)Math.Sin(p.rotation);
                    cos = (float)Math.Cos(p.rotation);
                    lastrotation = p.rotation;
                    p.position.X += cos * loc.X - sin * loc.Y;
                    p.position.Y += cos * loc.Y + sin * loc.X;
                }

                p.previousPosition.X = p.position.X;
                p.previousPosition.Y = p.position.Y;
            }
            else
            {
                p.previousPosition.X = p.position.X;
                p.previousPosition.Y = p.position.Y;

                p.position.X = 0.0f;
                p.position.Y = 0.0f;
            }
        }
    }
}
