﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="ScaleAll.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2008, 2009, 2010, 2011
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
//   This file is part of starLiGHT.Particles.
//
//   starLiGHT.Particles is free software: you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   starLiGHT.Particles is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU Lesser General Public License for more details.
//
//   You should have received a copy of the GNU Lesser General Public License
//   along with starLiGHT.Particles. If not, see http://www.gnu.org/licenses/.
//    
//   ADDITIONAL (commercial) LICENSES for starLiGHT.Particles are available on request.
// </license>
// <original_license>
//  * FLINT PARTICLE SYSTEM
//  * .....................
//  * 
//  * Author: Richard Lord
//  * Copyright (c) Richard Lord 2008-2009
//  * http://flintparticles.org
//  * 
//  * 
//  * Licence Agreement
//  * 
//  * Permission is hereby granted, free of charge, to any person obtaining a copy
//  * of this software and associated documentation files (the "Software"), to deal
//  * in the Software without restriction, including without limitation the rights
//  * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  * copies of the Software, and to permit persons to whom the Software is
//  * furnished to do so, subject to the following conditions:
//  * 
//  * The above copyright notice and this permission notice shall be included in
//  * all copies or substantial portions of the Software.
//  * 
//  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  * THE SOFTWARE.
// </original_license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 7773                  $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2011-03-28 02:04:04 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion

namespace starLiGHT.ParticleSystem.Initializer.TwoD
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using starLiGHT.ParticleSystem.Common;
    using starLiGHT.ParticleSystem.Initializer;

    #endregion

    /// <summary>
    /// The ScaleAll Initializer sets the size of the particle and adjusts 
    /// it's mass and collision radius accordingly.
    /// </summary>
    /// <remarks>
    /// <para>The scale factor of particles initialized by this initializer will be a 
    /// random value between the minimum and maximum values set. If no maximum value is 
    /// set, the minimum value is used with no variation.</para>
    /// <para>If you want to adjust only the size use the Scale Initializer.</para>
    /// <para>This initializer has a default priority of -10 to ensure it occurs after
    /// mass and radius assignment classes like CollisionRadius and Mass.</para>
    /// </remarks>
    public class ScaleAll : InitializerBase
    {
        #region Member Variables
        private float _min;
        private float _max;

        #endregion

        #region Constructors
        /// <summary>
        /// The constructor creates a ScaleAll Initializer to be used by an emitter.
        /// </summary>
        /// <remarks>
        /// <para>If you want to adjust only the size use the Scale Initializer.</para>
        /// <para>This initializer has a default priority of -10 to ensure it occurs after
        /// mass and radius assignment classes like CollisionRadius and Mass.</para>
        /// </remarks>
        public ScaleAll()
            : this(1.0f, float.NaN)
        {
        }

        /// <summary>
        /// The constructor creates a ScaleAll Initializer to be used by an emitter.
        /// </summary>
        /// <remarks>
        /// <param name="maxScale">
        /// The maximum scale factor for particles initialized by this class.
        /// </param>
        /// <param name="minScale">
        /// The minimum scale factor for particles initialized by this class.
        /// </param>
        /// <para>The scale factor of particles initialized by this initializer will be a 
        /// random value between the minimum and maximum values set. If no maxi,um value is 
        /// set, the minimum value is used with no variation.</para>
        /// <para>If you want to adjust only the size use the Scale Initializer.</para>
        /// <para>This initializer has a default priority of -10 to ensure it occurs after
        /// mass and radius assignment classes like CollisionRadius and Mass.</para>
        /// </remarks>
        public ScaleAll(float minScale, float maxScale)
        {
            Priority = -10;
            _min = minScale;
            _max = float.IsNaN(maxScale) ? minScale : maxScale;
        }

        /// <summary>
        /// The constructor creates a ScaleAll Initializer to be used by an emitter.
        /// </summary>
        /// <remarks>
        /// <para>If you want to adjust only the size use the Scale Initializer.</para>
        /// <para>This initializer has a default priority of -10 to ensure it occurs after
        /// mass and radius assignment classes like CollisionRadius and Mass.</para>
        /// </remarks>
        public ScaleAll(float scale)
        {
            Priority = -10;
            _min = _max = scale;
        }

        #endregion

        /// <summary>
        /// The minimum scale factor for particles initialized by this class.
        /// </summary>
        public float MinScale
        {
            get { return _min; }
            set { _min = value; }
        }

        /// <summary>
        /// The maximum scale factor for particles initialized by this class.
        /// </summary>
        public float MaxScale
        {
            get { return _max; }
            set { _max = value; }
        }

        /// <summary>
        /// When getting, returns the average of MinScale and MaxScale. When setting 
        /// this sets both MaxScale and MinScale to the same value.
        /// </summary>
        public float Scale
        {
            get { return _min == _max ? _min : (_max + _min) * 0.5f; }
            set { _max = _min = value; }
        }

        /// <summary>
        /// The initialize method is used by the emitter to apply the initialization
        /// to every particle. It is the key feature of the initializers and is
        /// used to initialize the state of every particle. This method 
        /// is called within the emitter's createParticle met6hod for every particle 
        /// and need not be called by the user.
        /// </summary>
        /// <param name="emitter">The Emitter that created the particle.</param>
        /// <param name="particle">The particle to be initialized.</param>
        /// <remarks>This is an internal method that is called by the emitter.</remarks>
        public override void initialize(Emitter emitter, ref Particle particle)
        {
            float scale = _max == _min ? _min : emitter.Random.NextFloat(_min, _max);
            particle.scale = scale;
            particle.mass *= scale * scale;
            particle.collisionRadius *= scale;
        }
    }
}
