﻿#region COPYRIGHT
//--------------------------------------------------------------------------------
// <copyright file="PrimitiveBatch.cs" company="starLiGHT Entertainment Studios">
//    Copyright (c) 2007, 2008, 2009, 2010
//       Roland Rosenkranz (Glatzemann@email.de)
// </copyright>
// <license>
// </license>
// <version>
// **************[ starLiGHT.Engine SVN ]**********************
// * $Rev:: 620                   $:  Revision of last commit *
// * $Author:: unknown            $:  Author of last commit   *
// * $Date:: 2010-12-07 15:23:58 #$:  Date of last commit     *
// ************************************************************
// </version>
//--------------------------------------------------------------------------------
#endregion
#if !ZUNE && !XNA4

namespace starLiGHT.Render
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using System.Text;
    using Microsoft.Xna.Framework;
    using Microsoft.Xna.Framework.Content;
    using Microsoft.Xna.Framework.Graphics;

    #endregion

    // PrimitiveBatch is a class that handles efficient rendering automatically for its
    // users, in a similar way to SpriteBatch. PrimitiveBatch can render lines, points,
    // and triangles to the screen.
    public class PrimitiveBatch : IDisposable
    {
#region Constants and Fields
        public float PointSize = 1.0f;
        public Matrix WorldMatrix = Matrix.Identity;
        public Matrix ViewMatrix = Matrix.Identity;
        public Matrix ProjectionMatrix = Matrix.Identity;

        // this constant controls how large the vertices buffer is. Larger buffers will
        // require flushing less often, which can increase performance. However, having
        // buffer that is unnecessarily large will waste memory.
        private const int DefaultBufferSize = 1500;

        // a block of vertices that calling AddVertex will fill. Flush will draw using
        // this array, and will determine how many primitives to draw from
        // this.positionInBuffer.
        private VertexPositionColor[] vertices = new VertexPositionColor[DefaultBufferSize];

        // keeps track of how many vertices have been added. this value increases until
        // we run out of space in the buffer, at which time Flush is automatically
        // called.
        private int positionInBuffer = 0;

        // the vertex declaration that will be set on the device for drawing. this is 
        // created automatically using VertexPositionColor's vertex elements.
        private VertexDeclaration vertexDeclaration;

        // a basic effect, which contains the shaders that we will use to draw our
        // primitives.
        private BasicEffect basicEffect;

        // the device that we will issue draw calls to.
        private GraphicsDevice device;

        // this value is set by Begin, and is the type of primitives that we are
        // drawing.
        private PrimitiveType primitiveType;

        // how many verts does each of these primitives take up? points are 1,
        // lines are 2, and triangles are 3.
        private int numVertsPerPrimitive;

        // this.hasBegun is flipped to true once Begin is called, and is used to make
        // sure users don't call End before Begin is called.
        private bool hasBegun = false;

        private bool isDisposed = false;

        #endregion

        // the constructor creates a new PrimitiveBatch and sets up all of the internals
        // that PrimitiveBatch will need.
        public PrimitiveBatch(GraphicsDevice graphicsDevice)
        {
            if (graphicsDevice == null)
            {
                throw new ArgumentNullException("graphicsDevice");
            }

            this.device = graphicsDevice;

            // create a vertex declaration, which tells the graphics card what kind of
            // data to expect during a draw call. We're drawing using
            // VertexPositionColors, so we'll use those vertex elements.
            this.vertexDeclaration = new VertexDeclaration(graphicsDevice, VertexPositionColor.VertexElements);

            // set up a new basic effect, and enable vertex colors.
            this.basicEffect = new BasicEffect(graphicsDevice, null);
            this.basicEffect.VertexColorEnabled = true;

            // projection uses CreateOrthographicOffCenter to create 2d projection
            // matrix with 0,0 in the upper left.
            this.basicEffect.Projection = Matrix.CreateOrthographicOffCenter(0, graphicsDevice.Viewport.Width, graphicsDevice.Viewport.Height, 0, 0, 1);
        }

        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (disposing && !this.isDisposed)
            {
                if (this.vertexDeclaration != null)
                {
                    this.vertexDeclaration.Dispose();
                }

                if (this.basicEffect != null)
                {
                    this.basicEffect.Dispose();
                }

                this.isDisposed = true;
            }
        }

        // Begin is called to tell the PrimitiveBatch what kind of primitives will be
        // drawn, and to prepare the graphics card to render those primitives.
        public void Begin(PrimitiveType primitiveType)
        {
            if (this.hasBegun)
            {
                throw new InvalidOperationException("End must be called before Begin can be called again.");
            }

            // these three types reuse vertices, so we can't flush properly without more
            // complex logic. Since that's a bit too complicated for this sample, we'll
            // simply disallow them.
            if (primitiveType == PrimitiveType.LineStrip ||
                primitiveType == PrimitiveType.TriangleStrip)
            {
                throw new NotSupportedException("The specified primitiveType is not supported by PrimitiveBatch.");
            }

            this.primitiveType = primitiveType;

            // how many verts will each of these primitives require?
            this.numVertsPerPrimitive = NumVertsPerPrimitive(primitiveType);

            // prepare the graphics device for drawing by setting the vertex declaration
            // and telling our basic effect to begin.
            this.device.VertexDeclaration = this.vertexDeclaration;

            // flip the error checking boolean. It's now ok to call AddVertex, Flush,
            // and End.
            this.hasBegun = true;
        }

        // AddVertex is called to add another vertex to be rendered. To draw a point,
        // AddVertex must be called once. for lines, twice, and for triangles 3 times.
        // this function can only be called once begin has been called.
        // if there is not enough room in the vertices buffer, Flush is called
        // automatically.
        public void AddVertex(Vector3 vertex, Color color)
        {
            if (!this.hasBegun)
            {
                throw new InvalidOperationException("Begin must be called before AddVertex can be called.");
            }

            // are we starting a new primitive? if so, and there will not be enough room
            // for a whole primitive, flush.
            bool newPrimitive = (this.positionInBuffer % this.numVertsPerPrimitive) == 0;

            if (newPrimitive && (this.positionInBuffer + this.numVertsPerPrimitive) >= this.vertices.Length)
            {
                this.Flush();
            }

            // once we know there's enough room, set the vertex in the buffer,
            // and increase position.
            this.vertices[this.positionInBuffer].Position = vertex;
            this.vertices[this.positionInBuffer].Color = color;

            this.positionInBuffer++;
        }

        public void AddVertex(Vector2 vertex, Color color)
        {
            this.AddVertex(new Vector3(vertex, 0), color);
        }

        // End is called once all the primitives have been drawn using AddVertex.
        // it will call Flush to actually submit the draw call to the graphics card, and
        // then tell the basic effect to end.
        public void End()
        {
            if (!this.hasBegun)
            {
                throw new InvalidOperationException("Begin must be called before End can be called.");
            }

            // Draw whatever the user wanted us to draw
            this.Flush();

            this.hasBegun = false;
        }

        // Flush is called to issue the draw call to the graphics card. Once the draw
        // call is made, this.positionInBuffer is reset, so that AddVertex can start over
        // at the beginning. End will call this to draw the primitives that the user
        // requested, and AddVertex will call this if there is not enough room in the
        // buffer.
        private void Flush()
        {
            if (!this.hasBegun)
            {
                throw new InvalidOperationException("Begin must be called before Flush can be called.");
            }

            // no work to do
            if (this.positionInBuffer == 0)
            {
                return;
            }

            float oldPointSize = this.device.RenderState.PointSize;
            this.device.RenderState.PointSize = this.PointSize;

            this.basicEffect.Begin();
            this.basicEffect.CurrentTechnique.Passes[0].Begin();

            this.basicEffect.View = this.ViewMatrix;
            this.basicEffect.World = this.WorldMatrix;
            this.basicEffect.Projection = this.ProjectionMatrix;

            // how many primitives will we draw?
            int primitiveCount;

            if (this.primitiveType == PrimitiveType.TriangleFan)
            {
                primitiveCount = this.positionInBuffer - 2;
            }
            else
            {
                primitiveCount = this.positionInBuffer / this.numVertsPerPrimitive;
            }

            CullMode old = this.device.RenderState.CullMode;
            this.device.RenderState.CullMode = CullMode.None;

            // submit the draw call to the graphics card
            this.device.DrawUserPrimitives<VertexPositionColor>(this.primitiveType, this.vertices, 0, primitiveCount);

            this.device.RenderState.CullMode = old;

            // now that we've drawn, it's ok to reset this.positionInBuffer back to zero,
            // and write over any vertices that may have been set previously.
            this.positionInBuffer = 0;

            // and then tell basic effect that we're done.
            this.basicEffect.CurrentTechnique.Passes[0].End();
            this.basicEffect.End();

            this.device.RenderState.PointSize = oldPointSize;
        }

#region Helper functions

        // NumVertsPerPrimitive is a boring helper function that tells how many vertices
        // it will take to draw each kind of primitive.
        private static int NumVertsPerPrimitive(PrimitiveType primitive)
        {
            int numVertsPerPrimitive;
            switch (primitive)
            {
                case PrimitiveType.PointList:
                case PrimitiveType.TriangleFan:
                    numVertsPerPrimitive = 1;
                    break;
                case PrimitiveType.LineList:
                    numVertsPerPrimitive = 2;
                    break;
                case PrimitiveType.TriangleList:
                    numVertsPerPrimitive = 3;
                    break;
                default:
                    throw new InvalidOperationException("primitive is not valid");
            }

            return numVertsPerPrimitive;
        }

#endregion
    }
}

#endif

#if XNA4
using System;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
namespace starLiGHT.Render
{
    // PrimitiveBatch is a class that handles efficient rendering automatically for its
    // users, in a similar way to SpriteBatch. PrimitiveBatch can render lines, points,
    // and triangles to the screen. In this sample, it is used to draw a spacewars
    // retro scene.
    public class PrimitiveBatch : IDisposable
    {
        #region Constants and Fields

        // this constant controls how large the vertices buffer is. Larger buffers will
        // require flushing less often, which can increase performance. However, having
        // buffer that is unnecessarily large will waste memory.
        const int DefaultBufferSize = 500;

        // a block of vertices that calling AddVertex will fill. Flush will draw using
        // this array, and will determine how many primitives to draw from
        // this.positionInBuffer.
        VertexPositionColor[] vertices = new VertexPositionColor[DefaultBufferSize];

        // keeps track of how many vertices have been added. this value increases until
        // we run out of space in the buffer, at which time Flush is automatically
        // called.
        int positionInBuffer = 0;

        // a basic effect, which contains the shaders that we will use to draw our
        // primitives.
        BasicEffect basicEffect;

        // the device that we will issue draw calls to.
        GraphicsDevice device;

        // this value is set by Begin, and is the type of primitives that we are
        // drawing.
        PrimitiveType primitiveType;

        // how many verts does each of these primitives take up? points are 1,
        // lines are 2, and triangles are 3.
        int numVertsPerPrimitive;

        // this.hasBegun is flipped to true once Begin is called, and is used to make
        // sure users don't call End before Begin is called.
        bool hasBegun = false;

        bool isDisposed = false;

        public Matrix WorldMatrix = Matrix.Identity;
        public Matrix ViewMatrix = Matrix.Identity;
        public Matrix ProjectionMatrix = Matrix.Identity;

        #endregion

        // the constructor creates a new PrimitiveBatch and sets up all of the internals
        // that PrimitiveBatch will need.
        public PrimitiveBatch(GraphicsDevice graphicsDevice)
        {
            if (graphicsDevice == null)
            {
                throw new ArgumentNullException("graphicsDevice");
            }
            this.device = graphicsDevice;

            // set up a new basic effect, and enable vertex colors.
            this.basicEffect = new BasicEffect(graphicsDevice);
            this.basicEffect.VertexColorEnabled = true;

            // projection uses CreateOrthographicOffCenter to create 2d projection
            // matrix with 0,0 in the upper left.
            this.basicEffect.Projection = Matrix.CreateOrthographicOffCenter
                (0, graphicsDevice.Viewport.Width,
                graphicsDevice.Viewport.Height, 0,
                0, 1);
        }

        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (disposing && !this.isDisposed)
            {
                if (this.basicEffect != null)
                    this.basicEffect.Dispose();

                this.isDisposed = true;
            }
        }

        // Begin is called to tell the PrimitiveBatch what kind of primitives will be
        // drawn, and to prepare the graphics card to render those primitives.
        public void Begin(PrimitiveType primitiveType)
        {
            if (this.hasBegun)
            {
                throw new InvalidOperationException
                    ("End must be called before Begin can be called again.");
            }

            // these three types reuse vertices, so we can't flush properly without more
            // complex logic. Since that's a bit too complicated for this sample, we'll
            // simply disallow them.
            if (primitiveType == PrimitiveType.LineStrip ||
                primitiveType == PrimitiveType.TriangleStrip)
            {
                throw new NotSupportedException
                    ("The specified primitiveType is not supported by PrimitiveBatch.");
            }

            this.primitiveType = primitiveType;

            // how many verts will each of these primitives require?
            this.numVertsPerPrimitive = NumVertsPerPrimitive(primitiveType);

            // flip the error checking boolean. It's now ok to call AddVertex, Flush,
            // and End.
            this.hasBegun = true;
        }

        // AddVertex is called to add another vertex to be rendered. To draw a point,
        // AddVertex must be called once. for lines, twice, and for triangles 3 times.
        // this function can only be called once begin has been called.
        // if there is not enough room in the vertices buffer, Flush is called
        // automatically.
        public void AddVertex(Vector2 vertex, Color color)
        {
            if (!this.hasBegun)
            {
                throw new InvalidOperationException
                    ("Begin must be called before AddVertex can be called.");
            }

            // are we starting a new primitive? if so, and there will not be enough room
            // for a whole primitive, flush.
            bool newPrimitive = ((this.positionInBuffer % this.numVertsPerPrimitive) == 0);

            if (newPrimitive &&
                (this.positionInBuffer + this.numVertsPerPrimitive) >= this.vertices.Length)
            {
                this.Flush();
            }

            // once we know there's enough room, set the vertex in the buffer,
            // and increase position.
            this.vertices[this.positionInBuffer].Position = new Vector3(vertex, 0);
            this.vertices[this.positionInBuffer].Color = color;

            this.positionInBuffer++;
        }

        public void AddVertex(Vector3 vertex, Color color)
        {
            this.AddVertex(new Vector2(vertex.X, vertex.Y), color);
        }

        // End is called once all the primitives have been drawn using AddVertex.
        // it will call Flush to actually submit the draw call to the graphics card, and
        // then tell the basic effect to end.
        public void End()
        {
            if (!this.hasBegun)
            {
                throw new InvalidOperationException
                    ("Begin must be called before End can be called.");
            }

            // Draw whatever the user wanted us to draw
            this.Flush();

            this.hasBegun = false;
        }

        // Flush is called to issue the draw call to the graphics card. Once the draw
        // call is made, this.positionInBuffer is reset, so that AddVertex can start over
        // at the beginning. End will call this to draw the primitives that the user
        // requested, and AddVertex will call this if there is not enough room in the
        // buffer.
        private void Flush()
        {
            if (!this.hasBegun)
            {
                throw new InvalidOperationException
                    ("Begin must be called before Flush can be called.");
            }

            // no work to do
            if (this.positionInBuffer == 0)
            {
                return;
            }

            // tell our basic effect to begin.
            this.basicEffect.CurrentTechnique.Passes[0].Apply();

            this.basicEffect.View = this.ViewMatrix;
            this.basicEffect.World = this.WorldMatrix;
            this.basicEffect.Projection = this.ProjectionMatrix;

            // how many primitives will we draw?
            int primitiveCount = this.positionInBuffer / this.numVertsPerPrimitive;

            // submit the draw call to the graphics card
            this.device.DrawUserPrimitives<VertexPositionColor>(this.primitiveType, this.vertices, 0,
                primitiveCount);

            // now that we've drawn, it's ok to reset this.positionInBuffer back to zero,
            // and write over any vertices that may have been set previously.
            this.positionInBuffer = 0;
        }

        #region Helper functions

        // NumVertsPerPrimitive is a boring helper function that tells how many vertices
        // it will take to draw each kind of primitive.
        static private int NumVertsPerPrimitive(PrimitiveType primitive)
        {
            int numVertsPerPrimitive;
            switch (primitive)
            {
                case PrimitiveType.LineList:
                    numVertsPerPrimitive = 2;
                    break;
                case PrimitiveType.TriangleList:
                    numVertsPerPrimitive = 3;
                    break;
                default:
                    throw new InvalidOperationException("primitive is not valid");
            }
            return numVertsPerPrimitive;
        }

        #endregion


    }

}
#endif