﻿#region Using Statements
using System;
using System.Collections;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using starLiGHT.Collision;
using starLiGHT.Collision.BroadPhase;
using starLiGHT.Collision.NarrowPhase;
using starLiGHT.Components.SandboxManager;
using starLiGHT.Collections;
using System.Diagnostics;
#endregion

namespace starLiGHT.Sandbox.Collision.Tests
{
    [SandboxTest("Breakout")]
    public sealed class Breakout : CollisionSystemTest
    {
        #region Member Variables
        private const int brickColumnCount = 18;
        private const int brickRowCount = 10;
        private const float brickWidth = 60f;
        private const float brickHeight = 34f;
        private const float brickOffset = 1f;

        private const float ballSpeed = 300f;

        private Microsoft.Xna.Framework.Game game;
        private bool debugDraw = true;
        private FastList<HullObject> bricks;

        private HullObject paddle;
        private HullObject ball;

        private Texture2D background;
        private Texture2D[] gemTextures;
        private Random randomNumber;
        #endregion

        public override void Initialize(Game1 game)
        {
            this.game = game;

            background = game.Content.Load<Texture2D>(@"Breakout/Background");

            CollisionManager = new starLiGHT.Collision.CollisionManager(BroadPhaseType.AxisAlignedRectangleBruteForce,
                                                                        NarrowPhaseType.SeparatingAxisTheorem,
                                                                        false, false, true);
            //CollisionManager = new starLiGHT.Collision.CollisionManager(BroadPhaseType.QuadTree,
            //                                                            NarrowPhaseType.SeparatingAxisTheorem,
            //                                                            true, false, false);
            CollisionManager.Collision += new EventHandler<CollisionEventArgs>(CollisionManager_Collision);

            // add the paddle
            paddle = new HullObject(new RectangleF(590f, 680f, 128f, 32f).Corners);
            paddle.Tag = "Paddle";
            paddle.Texture = game.Content.Load<Texture2D>("Breakout/paddle");
            CollisionManager.AddCollidable(paddle);

            // add the ball
            ball = new HullObject(new RectangleF(640f, 640f, 24f, 24f).Corners);
            ball.Tag = "Ball";
            ball.Velocity = new Vector2(0.5f, -0.5f);
            ball.Texture = game.Content.Load<Texture2D>("Breakout/ball");
            CollisionManager.AddCollidable(ball);

            gemTextures = new Texture2D[4];
            gemTextures[0] = game.Content.Load<Texture2D>("Breakout/gem_red");
            gemTextures[1] = game.Content.Load<Texture2D>("Breakout/gem_green");
            gemTextures[2] = game.Content.Load<Texture2D>("Breakout/gem_blue");
            gemTextures[3] = game.Content.Load<Texture2D>("Breakout/gem_purple");

            randomNumber = new Random();

            CreateBricks();

            CreateWalls();
        }

        void CollisionManager_Collision(object sender, CollisionEventArgs e)
        {
            if (e.Manifold.NarrowPhaseCollision)
            {
                ICollidable2D a;
                ICollidable2D b;

                if (e.Manifold.Object0.Tag.ToString() == "Ball")
                {
                    a = e.Manifold.Object0 as ICollidable2D;
                    b = e.Manifold.Object1 as ICollidable2D;
                }
                else if (e.Manifold.Object1.Tag.ToString() == "Ball")
                {
                    a = e.Manifold.Object1 as ICollidable2D;
                    b = e.Manifold.Object0 as ICollidable2D;
                }
                else
                {
                    return;
                }

                if (b.Tag.ToString() == "Brick")
                {
                    CollisionManager.RemoveCollidable(b);
                    bricks.Remove(b as HullObject);

                    Vector2 colNormal = e.Manifold.Separation.Value;
                    colNormal.Normalize();
                    if (colNormal.LengthSquared() > 0)
                    {
                        ball.Velocity = Vector2.Reflect(ball.Velocity, colNormal);
                        ball.Position -= e.Manifold.Separation.Value;
                    }
                }
                
                if (b.Tag.ToString() == "Paddle" || b.Tag.ToString() == "Wall")
                {
                    Vector2 colNormal = e.Manifold.Separation.Value;
                    colNormal.Normalize();
                    if (colNormal.LengthSquared() > 0)
                    {
                        ball.Velocity = Vector2.Reflect(ball.Velocity, colNormal);
                        ball.Position -= e.Manifold.Separation.Value;
                    }
                }
            }
        }

        public override void Update(GameTime gameTime, KeyboardState actState, KeyboardState lastState)
        {
            if (actState.IsKeyDown(Keys.D) && lastState.IsKeyUp(Keys.D))
            {
                debugDraw = !debugDraw;
            }

            if (!singleStep || (singleStep && actState.IsKeyDown(Keys.C) && lastState.IsKeyUp(Keys.C)))
            {
                Vector2 pos = paddle.Position;
                pos.X = MathHelper.Clamp(Mouse.GetState().X, 62f, 1090f);
                paddle.Position = pos;

                ball.Position += ball.Velocity * (float)gameTime.ElapsedGameTime.TotalSeconds * ballSpeed;
            }

            base.Update(gameTime, actState, lastState);
        }

        public override void Draw(GameTime gameTime, SpriteBatch spriteBatch)
        {
            spriteBatch.Begin();

            spriteBatch.Draw(background, Vector2.Zero, Color.White);

            foreach (HullObject brick in bricks)
            {
                spriteBatch.Draw(brick.Texture, brick.Position, Color.White);
            }
            spriteBatch.Draw(paddle.Texture, paddle.Position, Color.White);
            spriteBatch.Draw(ball.Texture, ball.Position, Color.White);

            if (debugDraw)
            {
                CollisionManager.DrawDebug(spriteBatch);
            }

            spriteBatch.End();
        }

        private void CreateBricks()
        {
            if (bricks == null)
            {
                bricks = new FastList<HullObject>();
            }

            // remove the "old" bricks from collision manager
            foreach (HullObject ho in bricks)
            {
                CollisionManager.RemoveCollidable(ho);
            }
            bricks.Clear();

            // add new bricks to the collision manager
            for (int y = 0; y < brickRowCount; y++)
            {
                for (int x = 0; x < brickColumnCount; x++)
                {
                    HullObject brick = new HullObject(new RectangleF(0f, 0f, brickWidth, brickHeight).Corners);
                    brick.Position = new Vector2(x * (brickWidth + brickOffset) + 62f + 30f, y * (brickHeight + brickOffset) + 73f + 30f);
                    //TODO: enable after debugging: 
                    brick.Static = true;
                    brick.Tag = "Brick";
                    brick.Texture = gemTextures[randomNumber.Next(0, 4)];
                    bricks.Add(brick);
                    CollisionManager.AddCollidable(brick);
                }
            }
        }

        private void CreateWalls()
        {
            HullObject wall = new HullObject(new RectangleF(0f, 0f, 62f, 720f).Corners);
            wall.Static = true;
            wall.Tag = "Wall";
            CollisionManager.AddCollidable(wall);

            wall = new HullObject(new RectangleF(1218f, 0f, 62f, 720f).Corners);
            wall.Static = true;
            wall.Tag = "Wall";
            CollisionManager.AddCollidable(wall);

            wall = new HullObject(new RectangleF(62f, 0f, 1156f, 73f).Corners);
            wall.Static = true;
            wall.Tag = "Wall";
            CollisionManager.AddCollidable(wall);
        }
    }
}
